from abc import ABC, abstractmethod
import ast
from typing import Any, Optional, Type, TypeVar

from .object_stream import ObjectStream, executor_attr_name
from .util_ast import function_call

T = TypeVar("T")


class EventDataset(ObjectStream[T], ABC):
    r"""
    A source of event objects (a ROOT file, or a xAOD file). This class
    should be sub-classed with the information about the actual dataset, and
    should never be created on its own.
    """

    def __init__(self, item_type: Type = Any):
        """
        Should not be called directly. Make sure to initialize this ctor
        or tracking information will be lost.
        """

        # Create the base AST node.
        ed_ast = function_call("EventDataset", [])

        # Safely store a reference to our executor in the AST in an attribute not used by the
        # the native Python ast module.
        setattr(ed_ast, executor_attr_name, self.execute_result_async)

        # Safely store a reference to this object
        setattr(ed_ast, "_eds_object", self)

        # Let ObjectStream take care of passing around this AST.
        super().__init__(ed_ast, item_type)

    def __repr__(self):
        return f"'{self.__class__.__name__}'"

    def __str__(self):
        return self.__class__.__name__

    @abstractmethod
    async def execute_result_async(self, a: ast.AST, title: Optional[str] = None) -> Any:
        """
        Override in your sub-class. The infrastructure will call this to render the result
        "locally", or as requested by the AST.
        """
        pass  # pragma: no cover


def find_EventDataset(a: ast.AST) -> ast.Call:
    r"""
    Given an input query ast, find the EventDataset `ast` and return it.

    Args:
        a:      An AST that represents a query

    Returns:
        The `EventDataset` `ast` at the root of this query. It will not be None.

    Exceptions:
        If there is more than one `EventDataset` found in the query or if there
        is no `EventDataset` at the root of the query, then an exception is thrown.
    """

    class ds_finder(ast.NodeVisitor):
        def __init__(self):
            self.ds: Optional[ast.Call] = None

        def visit_Call(self, node: ast.Call):
            if not isinstance(node.func, ast.Name):
                return self.generic_visit(node)
            if node.func.id != "EventDataset":
                return self.generic_visit(node)

            if self.ds is not None:
                raise Exception("AST Query has more than one EventDataset in it!")
            self.ds = node
            return node

    ds_f = ds_finder()
    ds_f.visit(a)

    if ds_f.ds is None:
        raise Exception(f"AST Query has no root EventDataset: {ast.dump(a)}")

    return ds_f.ds



# Test the object stream
import ast
import asyncio
import logging
from typing import Any, Iterable, Optional, Tuple, TypeVar

import pytest

from func_adl import EventDataset
from func_adl.object_stream import ObjectStream
from func_adl.type_based_replacement import func_adl_callback


class my_event(EventDataset):
    async def execute_result_async(self, a: ast.AST, title: Optional[str] = None):
        await asyncio.sleep(0.01)
        return a


class my_event_with_title(EventDataset):
    async def execute_result_async(self, a: ast.AST, title: Optional[str] = None):
        await asyncio.sleep(0.01)
        return a, title


class dd_jet:
    def pt(self) -> float:
        ...

    def eta(self) -> float:
        ...


T = TypeVar("T")


def add_md_for_type(s: ObjectStream[T], a: ast.Call) -> Tuple[ObjectStream[T], ast.Call]:
    return s.MetaData({"hi": "there"}), a


@func_adl_callback(add_md_for_type)
class dd_event:
    def Jets(self, bank: str) -> Iterable[dd_jet]:
        ...


class my_event_with_type(EventDataset[dd_event]):
    def __init__(self):
        super().__init__(dd_event)

    async def execute_result_async(self, a: ast.AST, title: Optional[str] = None):
        await asyncio.sleep(0.01)
        return a


class MyTestException(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)


class my_event_boom(EventDataset):
    async def execute_result_async(self, a: ast.AST, title: Optional[str]):
        await asyncio.sleep(0.01)
        raise MyTestException("this is a test bomb")


def test_simple_query():
    r = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value()
    )
    assert isinstance(r, ast.AST)


def test_two_simple_query():
    r1 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value()
    )
    r2 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsROOTTTree("junk.root", "analysis", ["jetPT"])
        .value()
    )

    assert ast.dump(r1) == ast.dump(r2)


def test_with_types():
    r1 = my_event_with_type().SelectMany(lambda e: e.Jets("jets"))
    r = r1.Select(lambda j: j.eta()).value()
    assert isinstance(r, ast.AST)
    assert "there" in ast.dump(r)


def test_simple_quer_with_title():
    r = (
        my_event_with_title()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value(title="onetwothree")
    )
    assert r[1] == "onetwothree"


def test_simple_query_lambda():
    r = (
        my_event()
        .SelectMany(lambda e: e.jets())
        .Select(lambda j: j.pT())
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value()
    )
    assert isinstance(r, ast.AST)


def test_simple_query_parquet():
    r = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsParquetFiles("junk.root", "jetPT")
        .value()
    )
    assert isinstance(r, ast.AST)


def test_two_simple_query_parquet():
    r1 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsParquetFiles("junk.root", "jetPT")
        .value()
    )
    r2 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsParquetFiles("junk.root", ["jetPT"])
        .value()
    )

    assert ast.dump(r1) == ast.dump(r2)


def test_simple_query_panda():
    r = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsPandasDF(["analysis", "jetPT"])
        .value()
    )
    assert isinstance(r, ast.AST)


def test_two_imple_query_panda():
    r1 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsPandasDF(["analysis"])
        .value()
    )
    r2 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsPandasDF(["analysis"])
        .value()
    )
    assert ast.dump(r1) == ast.dump(r2)


def test_simple_query_awkward():
    r = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsAwkwardArray(["analysis", "jetPT"])
        .value()
    )
    assert isinstance(r, ast.AST)


def test_simple_query_as_awkward():
    r = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .as_awkward(["analysis", "jetPT"])
        .value()
    )
    assert isinstance(r, ast.AST)


def test_two_similar_query_awkward():
    r1 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsAwkwardArray(["analysis"])
        .value()
    )
    r2 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsAwkwardArray("analysis")
        .value()
    )

    assert ast.dump(r1) == ast.dump(r2)


def test_metadata():
    r = (
        my_event()
        .MetaData({"one": "two", "two": "three"})
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .value()
    )
    assert isinstance(r, ast.AST)


def test_query_metadata():
    r = (
        my_event()
        .QMetaData({"one": "two", "two": "three"})
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .value()
    )
    assert isinstance(r, ast.AST)
    assert "MetaData" not in ast.dump(r)


def test_query_metadata_dup(caplog):
    r = (
        my_event()
        .QMetaData({"one": "two", "two": "three"})
        .QMetaData({"one": "two", "two": "three"})
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .value()
    )
    assert isinstance(r, ast.AST)
    assert len(caplog.text) == 0


def test_query_metadata_dup_update(caplog):
    caplog.set_level(logging.INFO)
    r = (
        my_event()
        .QMetaData({"one": "two", "two": "three"})
        .QMetaData({"one": "twoo", "two": "three"})
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value()
    )
    assert isinstance(r, ast.AST)
    assert "one" in caplog.text
    assert "twoo" in caplog.text
    assert "two" in caplog.text


def test_query_metadata_composable(caplog):
    caplog.set_level(logging.INFO)
    r_base = my_event().QMetaData({"one": "1"})

    # Each of these is a different base and should not interfear.
    r1 = r_base.QMetaData({"two": "2"})
    r2 = r_base.QMetaData({"two": "2+"})

    from func_adl.ast.meta_data import lookup_query_metadata

    assert lookup_query_metadata(r1, "two") == "2"
    assert lookup_query_metadata(r2, "two") == "2+"

    assert len(caplog.text) == 0


def test_nested_query_rendered_correctly():
    r = (
        my_event()
        .Where("lambda e: e.jets.Select(lambda j: j.pT()).Where(lambda j: j > 10).Count() > 0")
        .SelectMany("lambda e: e.jets()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value()
    )
    assert isinstance(r, ast.AST)
    assert "Select(source" not in ast.dump(r)


def test_query_with_comprehensions():
    r = (
        my_event()
        .Where("lambda e: [j.pT()>10 for j in e.jets].Count() > 0")
        .SelectMany("lambda e: e.jets()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value()
    )
    assert isinstance(r, ast.AST)
    assert "ListComp" not in ast.dump(r)


def test_bad_where():
    with pytest.raises(ValueError):
        my_event().Where("lambda e: 10").SelectMany("lambda e: e.jets()").AsROOTTTree(
            "junk.root", "analysis", "jetPT"
        ).value()


@pytest.mark.asyncio
async def test_await_exe_from_coroutine_with_throw():
    with pytest.raises(MyTestException):
        r = (
            my_event_boom()
            .SelectMany("lambda e: e.jets()")
            .Select("lambda j: j.pT()")
            .AsROOTTTree("junk.root", "analysis", "jetPT")
            .value_async()
        )
        await r


@pytest.mark.asyncio
async def test_await_exe_from_normal_function():
    r = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value_async()
    )
    assert isinstance(await r, ast.AST)


def test_ast_prop():
    r = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
    )

    assert isinstance(r.query_ast, ast.AST)
    assert isinstance(r.query_ast, ast.Call)


@pytest.mark.asyncio
async def test_2await_exe_from_coroutine():
    r1 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value_async()
    )
    r2 = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.eta()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value_async()
    )
    rpair = await asyncio.gather(r1, r2)
    assert isinstance(rpair[0], ast.AST)
    assert isinstance(rpair[1], ast.AST)


@pytest.mark.asyncio
async def test_passed_in_executor():
    logged_ast: Optional[ast.AST] = None

    async def my_exe(a: ast.AST, title: Optional[str]) -> Any:
        nonlocal logged_ast
        logged_ast = a
        return 1

    r = (
        my_event()
        .SelectMany("lambda e: e.jets()")
        .Select("lambda j: j.pT()")
        .AsROOTTTree("junk.root", "analysis", "jetPT")
        .value_async(executor=my_exe)
    )

    assert (await r) == 1
    assert logged_ast is not None


def test_untyped():
    r = my_event()
    assert r.item_type == Any


def test_typed():
    class Jet:
        def pt(self) -> float:
            ...

    class Evt:
        def Jets(self) -> Iterable[Jet]:
            ...

    class evt_typed(EventDataset[Evt]):
        def __init__(self):
            super().__init__(Evt)

        async def execute_result_async(self, a: ast.AST, title: Optional[str] = None):
            await asyncio.sleep(0.01)
            return a

    r = evt_typed()
    assert r.item_type is Evt


def test_typed_with_select():
    class Jet:
        ...

    class Evt:
        def Jets(self) -> Iterable[Jet]:
            ...

    class evt_typed(EventDataset[Evt]):
        def __init__(self):
            super().__init__(Evt)

        async def execute_result_async(self, a: ast.AST, title: Optional[str] = None):
            await asyncio.sleep(0.01)
            return a

    r = evt_typed().Select(lambda e: e.Jets())
    assert r.item_type is Iterable[Jet]


def test_typed_with_selectmany():
    class Jet:
        ...

    class Evt:
        def Jets(self) -> Iterable[Jet]:
            ...

    class evt_typed(EventDataset[Evt]):
        def __init__(self):
            super().__init__(Evt)

        async def execute_result_async(self, a: ast.AST, title: Optional[str] = None):
            await asyncio.sleep(0.01)
            return a

    r = evt_typed().SelectMany(lambda e: e.Jets())
    assert r.item_type is Jet


def test_typed_with_select_and_selectmany():
    class Jet:
        def pt(self) -> float:
            ...

    class Evt:
        def Jets(self) -> Iterable[Jet]:
            ...

    class evt_typed(EventDataset[Evt]):
        def __init__(self):
            super().__init__(Evt)

        async def execute_result_async(self, a: ast.AST, title: Optional[str] = None):
            await asyncio.sleep(0.01)
            return a

        def Jets(self) -> Iterable[Jet]:
            ...

    r1 = evt_typed().SelectMany(lambda e: e.Jets())
    r = r1.Select(lambda j: j.pt())
    assert r.item_type is float


def test_typed_with_where():
    class Evt:
        def MET(self) -> float:
            ...

    class evt_typed(EventDataset[Evt]):
        def __init__(self):
            super().__init__(Evt)

        async def execute_result_async(self, a: ast.AST, title: Optional[str] = None):
            await asyncio.sleep(0.01)
            return a

    r = evt_typed().Where(lambda e: e.MET() > 100)
    assert r.item_type is Evt


def test_typed_with_metadata():
    class Evt:
        ...

    class evt_typed(EventDataset[Evt]):
        def __init__(self):
            super().__init__(Evt)

        async def execute_result_async(self, a: ast.AST, title: Optional[str] = None):
            await asyncio.sleep(0.01)
            return a

    r = evt_typed().MetaData({})
    assert r.item_type is Evt



# flake8: noqa

# Dataset accessors
from .event_dataset import EventDataset, find_EventDataset

# Extra LINQ-like functions
from .functions import Range

# Main streaming object
from .object_stream import ObjectStream

# Decorators to add extra functionally to the module
from .type_based_replacement import (
    func_adl_callable,
    func_adl_callback,
    func_adl_parameterized_call,
    register_func_adl_os_collection,
)



import inspect
import sys
import typing
from typing import Any, Dict, Optional, Tuple, Type, TypeVar

if sys.version_info >= (3, 8):
    from typing import get_args, get_origin
else:  # pragma: no cover
    # TODO: Remove this when we drop support for 3.7
    def get_args(tp):
        "Return arguments - this is done very simply and will fail ugly"
        return getattr(tp, "__args__", ())

    def get_origin(tp):
        "Return the origin - this is done very simply and will fail ugly"
        return getattr(tp, "__origin__", None)


def is_iterable(t: Type) -> bool:
    "Is this type iterable?"
    while (t is not Any) and (not _is_iterable_direct(t)):
        t = get_inherited(t)

    return t is not Any


def _is_iterable_direct(t: Type) -> bool:
    "Is this type iterable?"
    return getattr(t, "_name", None) == "Iterable" or getattr(t, "__name__", None) == "Iterable"


def get_inherited(t: Type) -> Type:
    """Returns the inherited type of `t`

    Notes:
    * This works for 3.7 forward (but not back!)

    Args:
        t (Type): The type (a class) that we should look at

    Returns:
        Type: The type for an inherited class, or `Any` if none can be found
    """
    if hasattr(t, "__orig_bases__"):
        base_classes = getattr(t, "__orig_bases__", None)
    elif hasattr(t, "__origin__") and hasattr(t.__origin__, "__orig_bases__"):
        base_classes = t.__origin__.__orig_bases__
    else:
        return Any

    r = base_classes[0]  # type: ignore

    g_args = get_args(t)
    if len(g_args) > 0:
        mapping = {a.__name__: v for a, v in zip(r.__parameters__, g_args)}

        r_base = get_origin(r)
        assert r_base is not None, "Internal error"

        # Get us back to typing if this is a common interface.
        # This is not needed in python 3.11 and forward, where
        # collections.abc.X can are all be parameterized.
        if r_base.__name__ in typing.__dict__:
            r_base = typing.__dict__[r_base.__name__]

        # Re-parameterize the type with the information e have from this parameterization.
        r = r_base[tuple(_resolve_type(t_arg, mapping) for t_arg in get_args(r))]

    return r


def unwrap_iterable(t: Type) -> Type:
    "Unwrap an iterable type"
    # Try to find an iterable in the history somehow

    while (t is not Any) and (not _is_iterable_direct(t)):
        t = get_inherited(t)

    if t == Any:
        return Any

    a = get_args(t)
    assert len(a) == 1, f"Coding error - expected iterable type with a parameter, got {t}"
    return a[0]


def build_type_dict_from_type(t: Type, at_class: Optional[Type] = None) -> Dict[str, TypeVar]:
    """Build a dictionary of type variables from a type

    Args:
        t (Type): The type to build the dictionary from
        at_type (Type): If there is some inheritance, then drop down to this level.
                        if we can't find it, then fail badly.

    Returns:
        Dict[str, Type]: The dictionary of type variables
    """
    d = {}
    generic_type = get_origin(t)
    if generic_type is None:
        if at_class is not None:
            raise TypeError(f"Could not find type {str(at_class)} in {str(t)}")
        return {}

    if at_class is not None and generic_type is not at_class:
        try:
            return build_type_dict_from_type(get_inherited(t), at_class)
        except TypeError as e:
            raise TypeError(f"Looked for generic parameters in {str(t)}") from e

    for a in zip(generic_type.__parameters__, get_args(t)):
        d[a[0].__name__] = a[1]
    return d


def _resolve_type(t: Type, parameters: Dict[str, Type]) -> Optional[Type]:
    """Resolve any parameters in `t` with what we find in `parameters`

    int, {} => int
    ~T, {~T: int} => int
    ~K, {~T: int} => None

    Args:
        t (Type): The type to resolve
        parameters (Dict[str, Type]): The dict of types to resolve

    Returns:
        None if `t` is parameterized by unknown type var's
        The resolved type (a copy leaving `t` untouched) if TypeVar's are filled in
        The type if no substitution is required.
    """
    if isinstance(t, TypeVar):
        if t.__name__ in parameters:
            return parameters[t.__name__]
        return None

    template_params = getattr(t, "__parameters__", None)
    if template_params is not None and (len(template_params) > 0):
        resolved_params = [_resolve_type(p, parameters) for p in template_params]
        if None in resolved_params:
            return None
        return t[tuple(resolved_params)]

    # Non-parameterized types are easy
    return t


def resolve_type_vars(
    parameterized_type: Type, context_type: Type, at_class: Optional[Type] = None
) -> Optional[Type]:
    """Given an object definition in context_type, return a a resolved set of types based
    on parameterized type. If `parameterized_type` references unknown `TypeVars`, then return
    `None`.

    int, <any> -> int
    ~T, Iterable[T=int] -> int
    ~S, Iterable[T=int] -> None

    Args:
        parameterized_type (Type): The type whose parameters are to be resolved
        context_type (Type): The context to use to do the resolution
        at_class (Type): Where the resolution should happen

    Returns:
        Optional[Type]: [description]
    """
    try:
        s = build_type_dict_from_type(context_type, at_class)
    except TypeError:
        s = {}
    return _resolve_type(parameterized_type, s)


def get_class_name(t: Type) -> str:
    "Return a name of a class (parameterized if this is a generic)"
    n = getattr(t, "_name", None)
    if n is not None:
        return f'{n}[{",".join(get_class_name(a) for a in t.__args__)}]'

    n = getattr(t, "__name__", None)
    if n is not None:
        return n

    # bail because we don't know how to do it nicely here.
    return str(t)


def get_method_and_class(class_object: Type, method_name: str) -> Optional[Tuple[Type, Any]]:
    """Finds a method in the inheritance hierarchy of a class object, and returns the
    class object where the method is defined, and the method object. If there is no such
    method it returns None.

    Args:
        class_object (Type): The class object to start the method search
        method_name (str): Name of the method to search for

    Returns:
        Optional[Tuple[Type, Any]]: None if the method can't be found
            Type: The class object where the method is defined
            Any: The method object
    """
    # Safely handle crazy requests
    if class_object is Any:
        return None

    # Check for templated classes
    # TODO: Use inspect.getmro
    if not hasattr(class_object, "__mro__"):
        class_object = get_origin(class_object)

    # Walk the resolution hierarchy to find the method
    found_obj = None
    found_method = None
    for c in inspect.getmro(class_object):
        m = getattr(c, method_name, None)
        if found_method is None and m is None:
            # We can't find the method!
            return None
        if found_method is None:
            found_obj = c
            found_method = m
        else:
            if found_method == m:
                found_obj = c
            else:
                return (found_obj, found_method)

    return (found_obj, found_method)



# Need setuptools even though it isn't used - loads some plugins.
import os
from distutils.core import setup

from setuptools import find_packages  # noqa: F401

# Use the readme as the long description.
with open("README.md", "r") as fh:
    long_description = fh.read()


extras_require = {
    "test": [
        "pytest",
        "pytest-asyncio",
        "pytest-cov",
        "flake8",
        "coverage",
        "twine",
        "wheel",
        "astunparse",
        "black",
        "isort",
    ]
}
extras_require["complete"] = sorted(set(sum(extras_require.values(), [])))
extras_require["complete"] = sorted(set(sum(extras_require.values(), [])))

version = os.getenv("func_adl_version")
if version is None:
    import logging

    logging.error("func_adl_version environment variable not set")
    version = "0.0.1a1"
version = version.split("/")[-1]

setup(
    name="func_adl",
    version=version,
    packages=["func_adl", "func_adl/ast"],
    scripts=[],
    description="Functional Analysis Description Language Base Package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="G. Watts (IRIS-HEP/UW Seattle)",
    author_email="gwatts@uw.edu",
    maintainer="Gordon Watts (IRIS-HEP/UW Seattle)",
    maintainer_email="gwatts@uw.edu",
    url="https://github.com/iris-hep/func_adl",
    license="MIT",
    test_suite="tests",
    install_requires=["make-it-sync"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest>=3.9"],
    extras_require=extras_require,
    classifiers=[
        # "Development Status :: 3 - Alpha",
        # "Development Status :: 4 - Beta",
        "Development Status :: 5 - Production/Stable",
        # "Development Status :: 6 - Mature",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],
    data_files=[],
    python_requires=">=3.6, <3.12",
    platforms="Any",
)



import ast
import sys
from typing import Any, List

from func_adl.util_ast import lambda_build

if sys.version_info >= (3, 9):

    def unparse_ast(a: ast.AST) -> str:
        return ast.unparse(a)

else:  # pragma: no cover

    def unparse_ast(a: ast.AST) -> str:
        return ast.dump(a)


def resolve_syntatic_sugar(a: ast.AST) -> ast.AST:
    """Transforms python idioms into func_adl statements where it makes sense

    * List comprehensions are turned into `Select` statements
    * Generator expressions are turned into `Select` statements

    Args:
        a (ast.AST): The AST to scan for syntatic sugar

    Returns:
        ast.AST: The resolved syntax
    """

    class syntax_transformer(ast.NodeTransformer):
        def resolve_generator(
            self, lambda_body: ast.AST, generators: List[ast.comprehension], node: ast.AST
        ) -> ast.AST:
            """Translate a list comprehension or a generator to Select statements.

            `[j.pt() for j in jets] -> jets.Select(lambda j: j.pt())`

            Args:
                lambda_body (ast.AST): The target of the lambda expression
                generators (List[ast.comprehension]): The list of generators
                node (ast.AST): The original AST node

            Returns:
                ast.AST: The reformed ast (untouched if no expression detected)
            """
            a = node
            for c in reversed(generators):
                target = c.target
                if not isinstance(target, ast.Name):
                    raise ValueError(
                        f"Comprehension variable must be a name, but found {target}"
                        f" - {unparse_ast(node)}."
                    )
                if c.is_async:
                    raise ValueError(f"Comprehension can't be async - {unparse_ast(node)}.")
                source_collection = c.iter

                # Turn the if clauses into Where statements
                for a_if in c.ifs:
                    where_function = lambda_build(target.id, a_if)
                    source_collection = ast.Call(
                        func=ast.Attribute(attr="Where", value=source_collection, ctx=ast.Load()),
                        args=[where_function],
                        keywords=[],
                    )

                lambda_function = lambda_build(target.id, lambda_body)
                a = ast.Call(
                    func=ast.Attribute(attr="Select", value=source_collection, ctx=ast.Load()),
                    args=[lambda_function],
                    keywords=[],
                )

                # In case we have chained comprehensions
                lambda_body = a
            return a

        def visit_ListComp(self, node: ast.ListComp) -> Any:
            "Translate a list comprehension into a Select statement"
            a = self.generic_visit(node)

            if isinstance(a, ast.ListComp):
                a = self.resolve_generator(a.elt, a.generators, node)

            return a

        def visit_GeneratorExp(self, node: ast.GeneratorExp) -> Any:
            "Translate a generator into a Select statement"
            a = self.generic_visit(node)

            if isinstance(a, ast.GeneratorExp):
                a = self.resolve_generator(a.elt, a.generators, node)

            return a

    return syntax_transformer().visit(a)



import ast
from typing import Any, Dict, List, Optional, Tuple

from func_adl.ast.func_adl_ast_utils import FuncADLNodeTransformer
from func_adl.object_stream import ObjectStream


class _extract_metadata(FuncADLNodeTransformer):
    """Extract all the metadata from an expression, and remove the
    metadata nodes. Assume the metadata can all be bubbled to the top and
    has equal precedence.
    """

    def __init__(self):
        super().__init__()
        self._metadata = []

    @property
    def metadata(self) -> List[Dict[str, str]]:
        """Returns the metadata found while scanning expressions
        in the order it was encountered.

        Returns:
            List[Dict[str, str]]: List of all metadata found.
        """
        return self._metadata

    def visit_Call(self, node: ast.Call):
        """Detect a MetaData call, and remove it, storing the
        information.

        Args:
            node (ast.Call): The call node to process.

        Returns:
            ast.AST: The ast without the call node (if need be).
        """
        if isinstance(node.func, ast.Name) and node.func.id == "MetaData":
            self._metadata.append(ast.literal_eval(node.args[1]))
            return self.visit(node.args[0])
        return super().visit_Call(node)


def extract_metadata(a: ast.AST) -> Tuple[ast.AST, List[Dict[str, str]]]:
    """Returns the expresion with extracted metadata and the metadata, in order
    from the outter most to the inner most `MetaData` expressions.

    Args:
        a (ast.AST): The AST potentially containing metadata definitions

    Returns:
        Tuple[ast.AST, List[Dict[str, str]]]: a new AST without the metadata references
        and a list of metadata found.
    """
    e = _extract_metadata()
    a_new = e.visit(a)
    return a_new, e.metadata


def remove_empty_metadata(a: ast.AST) -> ast.AST:
    """Returns a new ast with any empty `MetaData` clauses removed.
    The old ast is not modified.

    Args:
        a (ast.AST): The AST of the query to clean up.

    Returns:
        ast.AST: The cleaned up AST.
    """

    class _cleaner(ast.NodeTransformer):
        def visit_Call(self, node: ast.Call):
            n = self.generic_visit(node)
            assert isinstance(n, ast.Call)
            if isinstance(n.func, ast.Name) and n.func.id == "MetaData":
                if len(n.args) == 2:
                    d = ast.literal_eval(n.args[1])
                    if isinstance(d, dict) and len(d) == 0:
                        return n.args[0]
            return n

    return _cleaner().visit(a)


def lookup_query_metadata(q: ObjectStream, metadata_name: str) -> Optional[Any]:
    """Walk back up the query tree to find metadata_name. Return `None` if not found.

    Args:
        q (ObjectStream): Object stream to walk back up.
        metadata_name (str): Name of the metadata item we are looking for

    Returns:
        Optional[Any]: Either the metadata value or `None`.
    """

    class _finder(ast.NodeVisitor):
        def __init__(self):
            self.ds: Optional[ast.Call] = None
            self._found = None

        @property
        def found(self) -> Optional[Any]:
            return self._found

        def generic_visit(self, node: ast.AST):
            q_metadata = getattr(node, "_q_metadata", None)
            found = False
            if q_metadata is not None:
                if metadata_name in q_metadata:
                    found = True
                    self._found = q_metadata[metadata_name]

            if not found:
                super().generic_visit(node)

    ds_f = _finder()
    ds_f.visit(q.query_ast)

    return ds_f.found
